# -*- coding: utf-8 -*-


from .compat import PY2, LOCALE, unicode, str_compat
from .handlers import commands as hc


@str_compat
class ProtocolError(IOError):

    def __str__(self):
        if self.errno is None and PY2:
            message = self.message
        else:
            # метод __str__ вернет str, если strerror является str
            # метод __str__ вернет unicode, если strerror является unicode
            message = super(ProtocolError, self).__str__()

        if PY2 and isinstance(message, str):
            message = message.decode(LOCALE)

        return message


class NoConnectionError(ProtocolError):
    
    def __init__(self, *args):
        if not args:
            args = (u'Нет связи с ККМ', )
        super(NoConnectionError, self).__init__(*args)


class UnexpectedResponseError(ProtocolError):
    pass


class FDError(ProtocolError):
    pass


@str_compat
class Error(ProtocolError):

    codes = {
        0x00: u'Ошибок нет',
        0x01: u'Неисправен накопитель ФП 1, ФП 2 или часы',
        0x02: u'Отсутствует ФП 1',
        0x03: u'Отсутствует ФП 2',
        0x04: u'Некорректные параметры в команде обращения к ФП',
        0x05: u'Нет запрошенных данных',
        0x06: u'ФП в режиме вывода данных',
        0x07: u'Некорректные параметры в команде для данной реализации ФП',
        0x08: u'Команда не поддерживается в данной реализации ФП',
        0x09: u'Некорректная длина команды',
        0x0A: u'Формат данных не BCD',
        0x0B: u'Неисправна ячейка памяти ФП при записи итога',
        0x0C: u'Переполнение необнуляемой суммы',
        0x0D: u'Переполнение суммы итогов смен',
        0x11: u'Не введена лицензия',
        0x12: u'Заводской номер уже введен',
        0x13: u'Текущая дата меньше даты последней записи в ФП',
        0x14: u'Область сменных итогов ФП переполнена',
        0x15: u'Смена уже открыта',
        0x16: u'Смена не открыта',
        0x17: u'Номер первой смены больше номера последней смены',
        0x18: u'Дата первой смены больше даты последней смены',
        0x19: u'Нет данных в ФП',
        0x1A: u'Область перерегистраций в ФП переполнена',
        0x1B: u'Заводской номер не введен',
        0x1C: u'В заданном диапазоне есть поврежденная запись',
        0x1D: u'Повреждена последняя запись сменных итогов',
        0x1E: u'Запись фискализации (перерегистрации ККМ) в накопителе не найдена',
        0x1F: u'Отсутствует память регистров',
        0x20: u'Переполнение денежного регистра при добавлении',
        0x21: u'Вычитаемая сумма больше содержимого денежного регистра',
        0x22: u'Неверная дата',
        0x23: u'Нет записи активизации',
        0x24: u'Область активизаций переполнена',
        0x25: u'Нет активизации с запрашиваемым номером',
        0x26: u'В ФП присутствует 3 или более битые записи сменных итогов.',
        0x27: u'Признак несовпадения КС, з/н, перерегистраций или активизаций',
        0x28: u'Технологическая метка в накопителе присутствует',
        0x29: u'Технологическая метка в накопителе отсутствует, возможно накопитель пуст',
        0x2A: u'Фактическая емкость микросхемы накопителя не соответствует текущей версии ПО',
        0x2B: u'Невозможно отменить предыдущую команду',
        0x2C: u'Обнулённая касса (повторное гашение невозможно)',
        0x2D: u'Сумма чека по секции меньше суммы сторно',
        0x2E: u'В ККТ нет денег для выплаты',
        0x2F: u'Не совпадает заводской номер ККМ в оперативной памяти ФП с номером в накопителе',
        0x30: u'ККТ заблокирован, ждет ввода пароля налогового инспектора',
        0x31: u'Сигнатура емкости накопителя не соответствует текущей версии ПО',
        0x32: u'Требуется выполнение общего гашения',
        0x33: u'Некорректные параметры в команде',
        0x34: u'Нет данных',
        0x35: u'Некорректный параметр при данных настройках',
        0x36: u'Некорректные параметры в команде для данной реализации ККТ',
        0x37: u'Команда не поддерживается в данной реализации ККТ',
        0x38: u'Ошибка в ПЗУ',
        0x39: u'Внутренняя ошибка ПО ККТ',
        0x3A: u'Переполнение накопления по надбавкам в смене',
        0x3B: u'Переполнение накопления в смене',
        0x3C: u'ЭКЛЗ: неверный регистрационный номер',
        0x3D: u'Смена не открыта – операция невозможна',
        0x3E: u'Переполнение накопления по секциям в смене',
        0x3F: u'Переполнение накопления по скидкам в смене',
        0x40: u'Переполнение диапазона скидок',
        0x41: u'Переполнение диапазона оплаты наличными',
        0x42: u'Переполнение диапазона оплаты типом 2',
        0x43: u'Переполнение диапазона оплаты типом 3',
        0x44: u'Переполнение диапазона оплаты типом 4',
        0x45: u'Сумма всех типов оплаты меньше итога чека',
        0x46: u'Не хватает наличности в кассе',
        0x47: u'Переполнение накопления по налогам в смене',
        0x48: u'Переполнение итога чека',
        0x49: u'Операция невозможна в открытом чеке данного типа',
        0x4A: u'Открыт чек – операция невозможна',
        0x4B: u'Буфер чека переполнен',
        0x4C: u'Переполнение накопления по обороту налогов в смене',
        0x4D: u'Вносимая безналичной оплатой сумма больше суммы чека',
        0x4E: u'Смена превысила 24 часа',
        0x4F: u'Неверный пароль',
        0x50: u'Идет печать результатов выполнения предыдущей команды',
        0x51: u'Переполнение накоплений наличными в смене',
        0x52: u'Переполнение накоплений по типу оплаты 2 в смене',
        0x53: u'Переполнение накоплений по типу оплаты 3 в смене',
        0x54: u'Переполнение накоплений по типу оплаты 4 в смене',
        0x55: u'Чек закрыт – операция невозможна',
        0x56: u'Нет документа для повтора',
        0x57: u'ЭКЛЗ: количество закрытых смен не совпадает с ФП',
        0x58: u'Ожидание команды продолжения печати',
        0x59: u'Документ открыт другим оператором',
        0x5A: u'Скидка превышает накопления в чеке',
        0x5B: u'Переполнение диапазона надбавок',
        0x5C: u'Понижено напряжение 24В',
        0x5D: u'Таблица не определена',
        0x5E: u'Неверная операция',
        0x5F: u'Отрицательный итог чека',
        0x60: u'Переполнение при умножении',
        0x61: u'Переполнение диапазона цены',
        0x62: u'Переполнение диапазона количества',
        0x63: u'Переполнение диапазона отдела',
        0x64: u'ФП отсутствует',
        0x65: u'Не хватает денег в секции',
        0x66: u'Переполнение денег в секции',
        0x67: u'Ошибка связи с ФП',
        0x68: u'Не хватает денег по обороту налогов',
        0x69: u'Переполнение денег по обороту налогов',
        0x6A: u'Ошибка питания в момент ответа по I2C',
        0x6B: u'Нет чековой ленты',
        0x6C: u'Нет контрольной ленты',
        0x6D: u'Не хватает денег по налогу',
        0x6E: u'Переполнение денег по налогу',
        0x6F: u'Переполнение по выплате в смене',
        0x70: u'Переполнение ФП',
        0x71: u'Ошибка отрезчика',
        0x72: u'Команда не поддерживается в данном подрежиме',
        0x73: u'Команда не поддерживается в данном режиме',
        0x74: u'Ошибка ОЗУ',
        0x75: u'Ошибка питания',
        0x76: u'Ошибка принтера: нет импульсов с тахогенератора',
        0x77: u'Ошибка принтера: нет сигнала с датчиков',
        0x78: u'Замена ПО',
        0x79: u'Замена ФП',
        0x7A: u'Поле не редактируется',
        0x7B: u'Ошибка оборудования',
        0x7C: u'Не совпадает дата',
        0x7D: u'Неверный формат даты',
        0x7E: u'Неверное значение в поле длины',
        0x7F: u'Переполнение диапазона итога чека',
        0x80: u'Ошибка связи с ФП (превышен таймаут I2C с контроллером)',
        0x81: u'Ошибка связи с ФП (контроллер отсутствует!? (получен NAK по I2C) '
              u'или принят неполный кадр от контроллера UART)',
        0x82: u'Ошибка связи с ФП (неверный формат данных в кадре I2C)',
        0x83: u'Ошибка связи с ФП (неверная контрольная сумма передаваемого кадра по I2C)',
        0x84: u'Переполнение наличности',
        0x85: u'Переполнение по продажам в смене',
        0x86: u'Переполнение по покупкам в смене',
        0x87: u'Переполнение по возвратам продаж в смене',
        0x88: u'Переполнение по возвратам покупок в смене',
        0x89: u'Переполнение по внесению в смене',
        0x8A: u'Переполнение по надбавкам в чеке',
        0x8B: u'Переполнение по скидкам в чеке',
        0x8C: u'Отрицательный итог надбавки в чеке',
        0x8D: u'Отрицательный итог скидки в чеке',
        0x8E: u'Нулевой итог чека',
        0x8F: u'Касса не фискализирована',
        0x90: u'Поле превышает размер, установленный в настройках',
        0x91: u'Выход за границу поля печати при данных настройках шрифта',
        0x92: u'Наложение полей',
        0x93: u'Восстановление ОЗУ прошло успешно',
        0x94: u'Исчерпан лимит операций в чеке',
        0x95: u'Неизвестная ошибка ЭКЛЗ',
        0x96: u'Выполните суточный отчет с гашением',
        0x9B: u'Некорректное действие',
        0x9C: u'Товар не найден по коду в базе товаров',
        0x9D: u'Неверные данные в записе о товаре в базе товаров',
        0x9E: u'Неверный размер файла базы или регистров товаров',
        0xA0: u'Ошибка связи с ЭКЛЗ',
        0xA1: u'ЭКЛЗ отсутствует',
        0xA2: u'ЭКЛЗ: Некорректный формат или параметр команды',
        0xA3: u'Некорректное состояние ЭКЛЗ',
        0xA4: u'Авария ЭКЛЗ',
        0xA5: u'Авария КС в составе ЭКЛЗ',
        0xA6: u'Исчерпан временной ресурс ЭКЛЗ',
        0xA7: u'ЭКЛЗ переполнена',
        0xA8: u'ЭКЛЗ: Неверные дата и время',
        0xA9: u'ЭКЛЗ: Нет запрошенных данных',
        0xAA: u'Переполнение ЭКЛЗ (отрицательный итог документа)',
        0xAF: u'Некорректные значения принятых данных от ЭКЛЗ',
        0xB0: u'ЭКЛЗ: Переполнение в параметре количество',
        0xB1: u'ЭКЛЗ: Переполнение в параметре сумма',
        0xB2: u'ЭКЛЗ: Уже активизирована',
        0xB4: u'Найденная запись фискализации (регистрации ККМ) повреждена',
        0xB5: u'Запись заводского номера ККМ повреждена',
        0xB6: u'Найденная запись активизации ЭКЛЗ повреждена',
        0xB7: u'Записи сменных итогов в накопителе не найдены',
        0xB8: u'Последняя запись сменных итогов не записана',
        0xB9: u'Сигнатура версии структуры данных в накопителе не совпадает с текущей версией ПО',
        0xBA: u'Структура накопителя повреждена',
        0xBB: u'Текущая дата+время меньше даты+времени последней записи активизации ЭКЛЗ',
        0xBC: u'Текущая дата+время меньше даты+времени последней записи фискализации (перерегистрации ККМ)',
        0xBD: u'Текущая дата меньше даты последней записи сменного итога',
        0xBE: u'Команда не поддерживается в текущем состоянии',
        0xBF: u'Инициализация накопителя невозможна',
        0xC0: u'Контроль даты и времени (подтвердите дату и время)',
        0xC1: u'ЭКЛЗ: суточный отчёт с гашением прервать нельзя',
        0xC2: u'Превышение напряжения в блоке питания',
        0xC3: u'Несовпадение итогов чека и ЭКЛЗ',
        0xC4: u'Несовпадение номеров смен',
        0xC5: u'Буфер подкладного документа пуст',
        0xC6: u'Подкладной документ отсутствует',
        0xC7: u'Поле не редактируется в данном режиме',
        0xC8: u'Отсутствуют импульсы от таходатчика',
        0xC9: u'Перегрев печатающей головки',
        0xCA: u'Температура вне условий эксплуатации',
        0xCB: u'Неверный подытог чека',
        0xCC: u'Смена в ЭКЛЗ уже закрыта',
        0xCD: u'Обратитесь в ЦТО: тест целостности архива ЭКЛЗ не прошел, '
              u'код ошибки ЭКЛЗ можно запросить командой 10H',
        0xCE: u'Лимит минимального свободного объема ОЗУ или ПЗУ на ККМ исчерпан',
        0xCF: u'Неверная дата (Часы сброшены? Установите дату!)',
        0xD0: u'Отчет по контрольной ленте не распечатан!',
        0xD1: u'Нет данных в буфере',
        0xD5: u'Критическая ошибка при загрузке ERRxx',
        0xE0: u'Ошибка связи с купюроприемником',
        0xE1: u'Купюроприемник занят',
        0xE2: u'Итог чека не соответствует итогу купюроприемника',
        0xE3: u'Ошибка купюроприемника',
        0xE4: u'Итог купюроприемника не нулевой'
    }

    fs_codes = {
        0x00: u'Успешное выполнение команды',
        0x01: u'Неизвестная команда, неверный формат посылки или неизвестные параметры',
        0x02: u'Неверное состояние ФН',
        0x03: u'Ошибка ФН',
        0x04: u'Ошибка КС',
        0x05: u'Закончен срок эксплуатации ФН',
        0x06: u'Архив ФН переполнен',
        0x07: u'Неверные дата и/или время',
        0x08: u'Нет запрошенных данных',
        0x09: u'Некорректное значение параметров команды',
        0x10: u'Превышение размеров TLV данных',
        0x11: u'Нет транспортного соединения',
        0x12: u'Исчерпан ресурс КС (криптографического сопроцессора)',
        0x14: u'Исчерпан ресурс хранения',
        0x15: u'Исчерпан ресурс Ожидания передачи сообщения',
        0x16: u'Продолжительность смены более 24 часов',
        0x17: u'Неверная разница во времени между 2 операцими',
        0x20: u'Сообщение от ОФД не может быть принято'
    }

    def __init__(self, cmd=None, code=None, message=None, fs=False):
        self.cmd = cmd or 0x00
        self.cmd_name = hc.COMMANDS.get(cmd, u'Неизвестная команда')
        self.fs = fs
        self.code = code or 0xFF

        if self.fs and self.code in self.fs_codes:
            self.code_desc = self.fs_codes[self.code]
        else:
            self.code_desc = self.codes.get(self.code, u'Неизвестная ошибка')

        if message:
            self.template = u'{message}'
            self.message = message if isinstance(message, unicode) else message.decode(LOCALE)
        else:
            self.template = u'0x{cmd:02X} ({cmd_name}) - {message} (0x{code:02X})'
            self.message = self.code_desc

    def __str__(self):
        return self.template.format(**self.__dict__)

    def __repr__(self):
        return '{}({})'.format(type(self).__name__, self)


class CheckError(Error):

    def __init__(self, exc):
        if isinstance(exc, Error):
            super(CheckError, self).__init__(cmd=exc.cmd, code=exc.code, fs=exc.fs)
        elif isinstance(exc, ProtocolError):
            super(CheckError, self).__init__(message=unicode(exc))
        else:
            raise ValueError(
                'Ожидался экземпляр класса {} или его подклассов, получен {}.'.format(
                    ProtocolError.__name__,
                    type(exc).__name__
                )
            )


class OpenCheckError(CheckError):
    pass


class ItemSaleError(CheckError):
    pass


class CloseCheckError(CheckError):
    pass



# -*- coding: utf-8 -*-


import struct
import operator
import functools
import collections

from .compat import reduce, xrange


NULL = bytearray((0x00, ))

DEFAULT_MIN_LENGTH = 40
T_TAPES = collections.namedtuple('Tapes', ['control', 'cash', 'skid'])

BAUDRATE_DIRECT = {
    2400: 0,
    4800: 1,
    9600: 2,
    19200: 3,
    38400: 4,
    57600: 5,
    115200: 6
}
BAUDRATE_REVERSE = {v: k for k, v in BAUDRATE_DIRECT.items()}

TLV_LEN_MAX = 250


def bytearray_cast(arg):
    if not isinstance(arg, bytearray):
        return bytearray(arg)
    return arg


def bytearray_concat(*args):
    """
    Функция конкатенирования нескольких bytearray в один.
    """

    return bytearray_cast(reduce(operator.concat, args))


def bytearray_strip(arg):
    """
    Функция отрезания нулевых байт от набора байт.
    """

    return arg.strip(NULL)


class mslice(object):
    def __init__(self, *args):
        """
        Класс множественных срезов.

        :param args: набор объектов класса slice
        """

        if not all(map(lambda i: isinstance(i, slice), args)):
            raise TypeError('support only slice objects')

        self.slices = args

    def __call__(self, arg):
        """
        Вызов.

        :type arg: bytearray
        :param arg: нарезаемый объект

        :rtype: bytearray
        :return: нарезанный объект
        """

        return bytearray_concat(*(arg[s] for s in self.slices))


def lrc(buff):
    """
    Расчет контрольной суммы.
    """

    return reduce(operator.xor, buff)


def encode(text):
    """
    Кодирование текста для передачи фискальному регистратору.
    """

    return text.encode('cp1251')


def decode(text):
    """
    Декодирование текста полученного с фискального регистратора.
    """

    return text.decode('cp1251')


def prepare_string(string, length=DEFAULT_MIN_LENGTH):
    """
    Подготовка строки к отправке в ККМ.

    :type string: unicode
    :param string: строка для передачи в ККМ
    :type length: int
    :param length: максимальная длина строки

    :rtype: bytearray
    :return: строка, готовая к передаче в ККМ
    """

    # нужно отправить в ККМ не менее DEFAULT_MIN_LENGTH символов
    if length < DEFAULT_MIN_LENGTH:
        length = DEFAULT_MIN_LENGTH

    if string:
        result = bytearray(encode(string)[:length])
        result.extend(NULL * (length - len(string)))
    else:
        result = NULL * length

    return result


def int_to_bits(num, length=None):
    """
    Функция, возвращающая битовое представление заданного числа.

    :type num: int
    :param num: целое число
    :type length: int
    :param length: количество бит, если указано, то число будет урезано до указанной длины

    :rtype: tuple
    :return: битовое представление числа
    """

    if not length:
        length = num.bit_length()

    result = [(num >> (1 * i) & 0x01) for i in xrange(length)]
    result.reverse()

    return tuple(result)


def bits_to_int(bits, reverse=False):
    """
    Функция, возвращающая целое число из набора бит.

    >>> bits_to_int((0, 0, 0, 1, 1, 0, 1, 1))
    27
    >>> bits_to_int((0, 0, 0, 1, 1, 0, 1, 1), True)
    216
    >>> bits_to_int((0, 0, 0, 0, 0, 0, 0, 1))
    1
    >>> bits_to_int((0, 0, 0, 0, 0, 0, 0, 1), True)
    128
    >>> bits_to_int((1, 0, 0, 0, 0, 0, 0, 1))
    129
    >>> bits_to_int((1, 0, 0, 0, 0, 0, 0, 1), True)
    129

    :type bits: collections.Iterable
    :param bits: набор бит
    :type reverse: bool
    :param reverse: интерпретировать биты в обратном порядке

    :type: int
    :return: целое число
    """

    result = 0

    if reverse:
        bits = reversed(bits)

    for bit in bits:
        result = (result << 1) | bit

    return result


def int_to_bytes(num, count=None):
    """
    Функция преобрабования целого числа в набор байт.

    :type num: int or long
    :param num: число
    :type count: int
    :param count: количество байт, если указано, то число будет усечено до
                  указанного количества байт

    :rtype: tuple
    :return: набор байт
    """

    if count:
        bytes_count = count
    else:
        if num == 0:
            bytes_count = 1
        else:
            q, r = divmod(num.bit_length(), 8)
            bytes_count = q + 1 if r else q

    return tuple((num >> (8 * i)) & 0xff for i in xrange(bytes_count))


def bytes_to_int(arg):
    """
    Функция преобразования набора байт в целое число.

    :type arg: list or tuple ot bytearray
    :param arg: набор байт

    :rtype: int
    :return: целое число
    """

    return sum(b << (8 * i) for i, b in enumerate(arg))


def chunks(source, num):
    """
    Функция нарезания списка на множество списков по заданному количеству элементов.

    >>> list(chunks([1, 0, 1, 1], 2)) == [[1, 0], [1, 1]]
    True

    :type source: collection.Iterable
    :param: source: исходный список
    :type num: int
    :param num: количество элементов

    :rtype: types.GeneratorType
    """

    for i in xrange(0, len(source), num):
        yield source[i:i + num]


def cast_byte_timeout(arg):
    """
    Приведение времени в секундах к коду таймаута.

    >>> cast_byte_timeout(0.0)
    0
    >>> cast_byte_timeout(0.15)
    150
    >>> cast_byte_timeout(0.30)
    151
    >>> cast_byte_timeout(15.0)
    249
    >>> cast_byte_timeout(30.0)
    250
    >>> cast_byte_timeout(105.0)
    255
    >>> cast_byte_timeout(0.16) # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
        ...
    ValueError: Неверное значение таймаута - 0.16 с.
    Значение должно соответсвовать одному из диапазонов: 0-0.15 с., 0.30-15 с., 30-105 с.
    >>> cast_byte_timeout(0.31) # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
        ...
    ValueError: Неверное значение таймаута - 0.31 с.
    Значение в диапазоне 0.30-15 с. должно быть кратно 0.15 с.
    >>> cast_byte_timeout(31.0) # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
        ...
    ValueError: Неверное значение таймаута - 31.0 с.
    Значение в диапазоне 30-105 с. должно быть кратно 15 с.

    :type arg: float
    :param arg: время в секундах

    :rtype: int
    :return: код таймаута
    """

    if 0.0 <= arg <= 0.15:
        return int(arg * 1000)
    elif 0.30 <= arg <= 15:
        div = arg / 0.15
        if not div.is_integer():
            raise ValueError(
                'Неверное значение таймаута - {} с. '
                'Значение в диапазоне 0.30-15 с. должно быть кратно 0.15 с.'.format(arg)
            )
        return int(div) + 149
    elif 30 <= arg <= 105:
        div = arg / 15
        if not div.is_integer():
            raise ValueError(
                'Неверное значение таймаута - {} с. '
                'Значение в диапазоне 30-105 с. должно быть кратно 15 с.'.format(arg)
            )
        return int(div) + 248
    else:
        raise ValueError(
            'Неверное значение таймаута - {} с. '
            'Значение должно соответсвовать одному из диапазонов: '
            '0-0.15 с., 0.30-15 с., 30-105 с.'.format(arg)
        )


class FuncChain(object):
    def __init__(self, *funcs):
        self.funcs = funcs

    def __call__(self, *args, **kwargs):
        if not self.funcs:
            return

        res = self.funcs[-1](*args, **kwargs)
        for func in reversed(self.funcs[:-1]):
            if res is None:
                break
            res = func(res)

        return res


fetchone = operator.itemgetter(0)


def unpack(fmt, string):
    """
    Функция, аналогичная функции struct.unpack, но возвращающая None в случае ошибки.
    """

    try:
        return struct.unpack(fmt, string)
    except struct.error:
        return


CMD_SIZE = {
    1: '>B',
    2: '>H'
}

CAST_CMD = {
    size: FuncChain(bytearray, functools.partial(struct.pack, fmt)) for size, fmt in CMD_SIZE.items()
}


CHAR_SIZE = {
    '1': '<B',
    '2': '<H',
    's2': '<h',
    '4': '<I',
    '11': '<2B',
    '111': '<3B',
    '1111': '<4B',
    '11111': '<5B',
    '121': '<BHB'
}

CAST_SIZE = {
    size: FuncChain(bytearray, functools.partial(struct.pack, fmt)) for size, fmt in CHAR_SIZE.items()
}

UNCAST_SIZE = {
    size: (
        FuncChain(fetchone, functools.partial(unpack, fmt))
        if len(size) == 1
        else functools.partial(struct.unpack, fmt)
    )
    for size, fmt in CHAR_SIZE.items()
}



# -*- coding: utf-8 -*-


import datetime
import binascii
import itertools

from .. import misc
from ..compat import PY2, str_compat


def handle_version(arg):
    return '.'.join(map(chr, arg))


def handle_date(arg):
    d, m, y = arg
    return datetime.date(2000 + y, m, d)


def handle_revdate(arg):
    try:
        res = datetime.date(arg[0] + 2000, *arg[1:])
    except ValueError:
        res = datetime.date.fromtimestamp(0)

    return res


def handle_time(arg):
    return datetime.time(*arg)


def handle_datetime(arg):
    try:
        res = datetime.datetime(arg[0] + 2000, *arg[1:])
    except ValueError:
        res = datetime.datetime.fromtimestamp(0)

    return res


def handle_fr_flags(arg):
    def get_keys(revision):
        return (
            (u'Увеличенная точность количества', u'Буфер принтера непуст')[revision],
            u'ЭКЛЗ почти заполнена',
            (u'Отказ левого датчика принтера', u'Бумага на выходе из презентера')[revision],
            (u'Отказ правого датчика принтера', u'Бумага на входе в презентер', u'Модель принтера')[revision],
            u'Денежный ящик',
            u'Крышка корпуса ФР',
            u'Рычаг термоголовки чековой ленты',
            u'Рычаг термоголовки контрольной ленты',
            u'Оптический датчик чековой ленты',
            u'Оптический датчик операционного журнала',
            u'ЭКЛЗ',
            u'Положение десятичной точки',
            u'Нижний датчик подкладного документа',
            u'Верхний датчик подкладного документа',
            u'Рулон чековой ленты',
            u'Рулон операционного журнала'
        )

    bits = misc.int_to_bits(arg, 16)

    a, b, c = 0, 1, 2
    flags_actual = {
        # ШТРИХ-ФР-К
        4: ((0, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1), a),
        # ШТРИХ-КОМБО-ФР-К
        9: ((0, 1, 0, 0, 1, 0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 0), a),
        # ШТРИХ-КОМБО-ФР-К (версия 02)
        12: ((0, 1, 0, 0, 1, 0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 0), a)
    }

    flags, rev = flags_actual.get(
        handle_fr_flags.model,
        ((1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1), a)
    )

    return dict(
        zip(
            itertools.compress(get_keys(rev), flags),
            itertools.compress(bits, flags)
        )
    )
handle_fr_flags.model = -1


def handle_baudrate(arg):
    """
    Функция обработки значения скорости обмена.
    """

    return misc.BAUDRATE_REVERSE.get(arg, 0)


def handle_byte_timeout(arg):
    """
    Функция обработки тайм аута приема байта.

    :type arg: int
    :param arg: код таймаута

    :rtype: float
    :return: время в секундах
    """

    if 0 <= arg <= 150:
        return arg * 0.001
    elif 151 <= arg <= 249:
        return (arg - 149) * 0.15
    elif 250 <= arg <= 255:
        return (arg - 248) * 15.0
    else:
        return -1


def handle_type_field(arg):
    """
    Функция обработки типа поля таблицы.
    """

    values = {
        0: int,
        1: str
    }

    return values[arg]


def handle_min_max_field_value(arg):
    """
    Функция обработки минимального и максимального значения для команды
    0x2E "Запрос структуры поля".
    """

    result = {}

    bytes_count = misc.UNCAST_SIZE['1'](arg[slice(0, 1)])
    result[u'Количество байт'] = bytes_count
    result[u'Минимальное значение поля'] = misc.bytes_to_int(
        arg[slice(1, bytes_count + 1)]
    )
    result[u'Максимальное значение поля'] = misc.bytes_to_int(
        arg[slice(bytes_count + 1, bytes_count * 2 + 1)]
    )

    return result


@str_compat
class FRMode(object):
    MODE_DESCR = {
        0: u'Принтер в рабочем режиме.',
        1: u'Выдача данных.',
        2: u'Открытая смена, 24 часа не кончились.',
        3: u'Открытая смена, 24 часа кончились.',
        4: u'Закрытая смена.',
        5: u'Блокировка по неправильному паролю налогового инспектора.',
        6: u'Ожидание подтверждения ввода даты.',
        7: u'Разрешение изменения положения десятичной точки.',
        8: u'Открытый документ.',
        9: u'Режим разрешения технологического обнуления.',
        10: u'Тестовый прогон.',
        11: u'Печать полного фискального отчета.',
        12: u'Печать отчёта ЭКЛЗ.',
        13: u'Работа с фискальным подкладным документом.',
        14: u'Печать подкладного документа.',
        15: u'Фискальный подкладной документ сформирован.'
    }
    MODE_STATUS = {
        8: {
            0: u'Продажа.',
            1: u'Покупка.',
            2: u'Возврат продажи.',
            3: u'Возврат покупки.'
        },
        13: {
            0: u'Продажа (открыт).',
            1: u'Покупка (открыт).',
            2: u'Возврат продажи (открыт).',
            3: u'Возврат покупки (открыт).'
        },
        14: {
            0: u'Ожидание загрузки.',
            1: u'Загрузка и позиционирование.',
            2: u'Позиционирование.',
            3: u'Печать.',
            4: u'Печать закончена.',
            5: u'Выброс документа.',
            6: u'Ожидание извлечения.',
        }
    }

    def __init__(self, code):
        self.num = code & 0x0f
        self.status = code >> 4

        self.msg = self.MODE_DESCR.get(self.num, u'Неизвестный режим.')
        if self.num in self.MODE_STATUS:
            self.msg = u'{} {}'.format(
                self.msg.replace(u'.', u':'),
                self.MODE_STATUS[self.num].get(self.status, u'Неизвестный статус режима.')
            )

    @property
    def state(self):
        return self.num, self.status

    def __str__(self):
        return self.msg

    __repr__ = __str__


@str_compat
class FRSubMode(object):
    SUBMODE_DESCR = {
        0: u'Бумага есть.',
        1: u'Нет бумаги.',
        2: u'ФР ждет бумагу для продолжения печати.',
        3: u'ФР ждет команду продолжения печати.',
        4: u'Печать фискальных отчетов.',
        5: u'Печать операции.'
    }

    def __init__(self, code):
        self.num = code
        self.msg = self.SUBMODE_DESCR[self.num]

    @property
    def state(self):
        return self.num

    def __str__(self):
        return self.msg

    __repr__ = __str__


def handle_fp_flags(arg):
    keys = (u'ФП 1', u'ФП 2', u'Лицензия', u'Переполнение ФП',
            u'Батарея ФП', u'Последняя запись ФП', u'Смена в ФП', u'24 часа в ФП')
    bits = misc.int_to_bits(arg, 8)

    return dict(zip(keys, bits))


def handle_inn(arg):
    inn = binascii.hexlify(arg)
    if not PY2:
        inn = inn.decode()

    if inn == 'ffffffffffff':
        return -1
    return int(inn, base=16)


def handle_fs_lifestate(arg):
    values = (
        u'Закончена передача фискальных данных в ОФД',
        u'Закрыт фискальный режим',
        u'Открыт фискальный режим',
        u'Проведена настройка ФН'
    )

    return dict(zip(values, misc.int_to_bits(arg, 4)))


def handle_fs_current_document(arg):
    values = {
        0x00: u'нет открытого документа',
        0x01: u'отчет о фискализации',
        0x02: u'отчет об открытии смены',
        0x04: u'кассовый чек',
        0x08: u'отчет о закрытии смены',
        0x10: u'отчет о закрытии фискального режима',
        0x11: u'бланк строкой отчетности',
        0x12: u'отчет об изменении параметров регистрации ККТ в связи с заменой ФН',
        0x13: u'отчет об изменении параметров регистрации ККТ',
        0x14: u'кассовый чек коррекции',
        0x15: u'БСО коррекции',
        0x17: u'отчет о текущем состоянии расчетов'
    }

    return values.get(arg, u'неизвестный тип документа')


def handle_fs_document_data(arg):
    values = {
        0: u'нет данных документа',
        1: u'получены данные документа'
    }

    return values[arg]


def handle_fs_shift_state(arg):
    values = {
        0: u'смена закрыта',
        1: u'смена открыта'
    }

    return values[arg]


def handle_fs_warning_flags(arg):
    values = (
        u'Превышено время ожидания ответа ОФД',
        u'Переполнение памяти ФН (Архив ФН заполнен на 90%)',
        u'Исчерпание ресурса криптографического сопроцессора (до окончания срока действия 30 дней)',
        u'Срочная замена криптографического сопроцессора (до окончания срока действия 3 дня)'
    )

    return dict(zip(values, misc.int_to_bits(arg, 4)))


def handle_info_exchange_state(arg):
    values = (
        u'Ожидание ответа на команду от ОФД',
        u'Изменились настройки соединения с ОФД',
        u'Есть команда от ОФД',
        u'Ожидание ответного сообщения (квитанции) от ОФД',
        u'Есть сообщение для передачи в ОФД',
        u'Транспортное соединение установлено'
    )

    return dict(zip(values, misc.int_to_bits(arg, 6)))



# -*- coding: utf-8 -*-


from . import misc, excepts
from .compat import unicode, bool_compat


@bool_compat
class FD(object):
    TAGS = {
        # тэг: (тип значения, признак обязательности соблюдения длины, максимальная длина)
        # телефон или электронный адрес покупателя
        1008: (unicode, False, 64)
    }

    CAST = {
        unicode: lambda x: x.encode('cp866')
    }
    LEN = {
        bytes: (len, lambda value, len_: value.ljust(len_))
    }

    def __init__(self, tags=None):
        """
        Структура для работы с фискальными данными.
        
        :type tags: dict
        :param tags: словарь {тэг: значение}
        """

        self.data = {}
        self.b_data = bytearray()

        tags = tags or {}
        for item in tags.items():
            self.set_value(*item)

    def set_value(self, tag, value):
        """
        Установить значение для тэга.
        
        :type tag: int
        :param tag: тэг
        :param value: значение тэга
        """

        try:
            type_, len_req, len_max = self.TAGS.get(tag)
        except TypeError:
            raise excepts.FDError(u'Тэг {} не поддерживается'.format(tag))

        value_type = type(value)
        if value_type != type_:
            raise excepts.FDError(
                u'Значение для тэга {} должно быть {}, получено {}'.format(tag, type_, value_type)
            )

        cast_call = self.CAST.get(value_type)
        if cast_call:
            value = cast_call(value)
            value_type = type(value)

        len_call, fill_call = self.LEN[value_type]
        if len_call(value) > len_max:
            raise excepts.FDError(u'Тэг {} имеет ограничение длины - {} байта'.format(tag, len_max))
        if len_req:
            value = fill_call(value, len_max)

        value_len = len_call(value)
        if not value_len:
            return

        self.data[tag] = value
        self.b_data.extend(
            misc.bytearray_concat(
                misc.CAST_SIZE['2'](tag),
                misc.CAST_SIZE['2'](len_call(value)),
                value
            )
        )

    def dump(self):
        """
        Получить TVL структуру, готовую для передачи в команду send_tlv_struct.
        
        :rtype: bytes
        :return: tlv строка
        """

        return bytes(self.b_data)

    def __bool__(self):
        return bool(self.data)



# -*- coding: utf-8 -*-


import sys
import locale


PY2 = sys.version_info.major == 2
LOCALE = locale.getpreferredencoding()


if PY2:
    unicode = unicode
    xrange = xrange
    reduce = reduce
else:
    import functools

    unicode = str
    xrange = range
    reduce = functools.reduce


# Взято из six
def with_metaclass(meta, *bases):
    """Create a base class with a metaclass."""
    # This requires a bit of explanation: the basic idea is to make a
    # dummy metaclass for one level of class instantiation that replaces
    # itself with the actual metaclass.
    class metaclass(type):
        def __new__(cls, name, this_bases, d):
            return meta(name, bases, d)

    return type.__new__(metaclass, 'temporary_class', (), {})


def str_compat(cls):
    if PY2:
        cls.__unicode__ = cls.__str__
        cls.__str__ = lambda self: self.__unicode__().encode(LOCALE)
        # перепривязываем __repr__
        if hasattr(cls, '__repr__') and cls.__repr__ == cls.__unicode__:
            cls.__repr__ = cls.__str__

    return cls


def bool_compat(cls):
    if PY2:
        cls.__nonzero__ = cls.__bool__

    return cls



# -*- coding: utf-8 -*-


import sys
import time
import inspect

from . import misc, excepts


def state(self):
    """
    Состояние ККМ в коротком виде.
    """

    return self.protocol.command(
        0x10,
        self.password
    )
state.cmd = 0x10


def full_state(self):
    """
    Состояние ККМ.
    """

    return self.protocol.command(
        0x11,
        self.password
    )
full_state.cmd = 0x11


def beep(self):
    """
    Гудок.
    """

    return self.protocol.command(
        0x13,
        self.password
    )
beep.cmd = 0x13


def set_exchange_params(self, port, baudrate, timeout):
    """
    Установка параметров обмена.
    """

    return self.protocol.command(
        0x14,
        self.admin_password,
        misc.CAST_SIZE['1'](port),
        misc.CAST_SIZE['1'](misc.BAUDRATE_DIRECT[baudrate]),
        misc.CAST_SIZE['1'](misc.cast_byte_timeout(timeout))
    )
set_exchange_params.cmd = 0x14


def read_exchange_params(self, port):
    """
    Чтение параметров обмена.
    """

    return self.protocol.command(
        0x15,
        self.admin_password,
        misc.CAST_SIZE['1'](port)
    )
read_exchange_params.cmd = 0x15


def reset_settings(self):
    """
    Технологическое обнуление.
    """

    return self.protocol.command_nopass(
        0x16
    )
reset_settings.cmd = 0x16


def print_string(self, string, control_tape=True, cash_tape=True):
    """
    Печать строки.
    """

    control = 0b01 if control_tape else 0b00
    cash = 0b10 if cash_tape else 0b00

    self.wait_printing()
    return self.protocol.command(
        0x17,
        self.password,
        misc.CAST_SIZE['1'](control + cash),
        misc.prepare_string(string, self.DEFAULT_MAX_LENGTH)
    )
print_string.cmd = 0x17


def print_line(self, symbol='-', control_tape=True, cash_tape=True):
    """
    Печать строки-разделителя.
    """

    return self.print_string(symbol * self.DEFAULT_MAX_LENGTH, control_tape, cash_tape)
print_string.related = (print_line, )


def test_start(self, minute):
    """
    Тестовый прогон.
    """

    self.wait_printing()
    return self.protocol.command(
        0x19,
        self.password,
        misc.CAST_SIZE['1'](minute)
    )
test_start.cmd = 0x19


def request_monetary_register(self, num):
    """
    Запрос денежного регистра.
    """

    return self.protocol.command(
        0x1A,
        self.password,
        misc.CAST_SIZE['1'](num)
    )
request_monetary_register.cmd = 0x1A


def request_operational_register(self, num):
    """
    Запрос операционного регистра.
    """

    return self.protocol.command(
        0x1B,
        self.password,
        misc.CAST_SIZE['1'](num)
    )
request_operational_register.cmd = 0x1B


def write_table(self, table, row, field, value, _type):
    """
    Запись таблицы.
    """

    cast_funcs_map = {
        int: misc.FuncChain(bytearray, misc.int_to_bytes),
        str: misc.encode
    }

    return self.protocol.command(
        0x1E,
        self.admin_password,
        misc.CAST_SIZE['121'](table, row, field),
        cast_funcs_map[_type](value)
    )
write_table.cmd = 0x1E


def read_table(self, table, row, field, _type):
    """
    Чтение таблицы.
    """

    cast_funcs_map = {
        int: misc.bytes_to_int,
        str: misc.FuncChain(misc.decode, misc.bytearray_strip)
    }

    if _type not in (cast_funcs_map.keys()):
        raise ValueError(
            u'ожидаемые типы {}'.format(', '.join(cast_funcs_map.keys()))
        )

    result = self.protocol.command(
        0x1F,
        self.admin_password,
        misc.CAST_SIZE['121'](table, row, field)
    )
    result[u'Значение'] = cast_funcs_map[_type](result[u'Значение'])

    return result
read_table.cmd = 0x1F


def set_time(self, time_):
    """
    Программирование времени.
    """

    # TODO: разобраться с округлением секунд до 00
    return self.protocol.command(
        0x21,
        self.admin_password,
        misc.CAST_SIZE['111'](time_.hour, time_.minute, time_.second)
    )
set_time.cmd = 0x21


def set_date(self, date):
    """
    Программирование даты.
    """

    return self.protocol.command(
        0x22,
        self.admin_password,
        misc.CAST_SIZE['111'](date.day, date.month, date.year - 2000)
    )
set_date.cmd = 0x22


def confirm_date(self, date):
    """
    Подтверждение программирование даты.
    """

    return self.protocol.command(
        0x23,
        self.admin_password,
        misc.CAST_SIZE['111'](date.day, date.month, date.year - 2000)
    )
confirm_date.cmd = 0x23


def set_datetime(self, datetime):
    """
    Установка даты и времени.
    """

    self.set_time(datetime.time())
    self.set_date(datetime.date())
    self.confirm_date(datetime.date())
set_time.related = (set_datetime, )
set_date.related = (set_datetime, )
confirm_date.related = (set_datetime, )
set_datetime.required = (set_time, set_date, confirm_date)


def cut(self, partial=False):
    """
    Обрезка чека.
    """

    self.wait_printing()
    return self.protocol.command(
        0x25,
        self.password,
        misc.CAST_SIZE['1'](partial)
    )
cut.cmd = 0x25


def open_drawer(self, box=0):
    """
    Открыть денежный ящик.
    """

    return self.protocol.command(
        0x28,
        self.password,
        misc.CAST_SIZE['1'](box)
    )
open_drawer.cmd = 0x28


def feed(self, count, control_tape=False, cash_tape=False, skid_document=False):
    """
    Протяжка чековой ленты на заданное количество строк.
    """

    if count > 255:
        raise ValueError(u'Количество строк должно быть меньше 255')

    control = 0b001 if control_tape else 0b000
    cash = 0b010 if cash_tape else 0b000
    skid = 0b100 if skid_document else 0b000

    self.wait_printing()
    return self.protocol.command(
        0x29,
        self.password,
        misc.CAST_SIZE['1'](control + cash + skid),
        misc.CAST_SIZE['1'](count)
    )
feed.cmd = 0x29


def test_stop(self):
    """
    Прерывание тестового прогона.
    """

    self.wait_printing()
    return self.protocol.command(
        0x2B,
        self.password
    )
test_stop.cmd = 0x2B


def request_table_structure(self, table):
    """
    Запрос структуры таблицы.
    """

    return self.protocol.command(
        0x2D,
        self.admin_password,
        misc.CAST_SIZE['1'](table)
    )
request_table_structure.cmd = 0x2D


def request_field_structure(self, table, field):
    """
    Запрос структуры поля.
    """

    return self.protocol.command(
        0x2E,
        self.admin_password,
        misc.CAST_SIZE['11'](table, field)
    )
request_field_structure.cmd = 0x2E


def x_report(self):
    """
    Суточный отчет без гашения.
    """

    self.wait_printing()
    return self.protocol.command(
        0x40,
        self.admin_password
    )
x_report.cmd = 0x40


def z_report(self):
    """
    Суточный отчет с гашением.
    """

    self.wait_printing()
    return self.protocol.command(
        0x41,
        self.admin_password
    )
z_report.cmd = 0x41


def income(self, cash):
    """
    Внесение.
    """

    self.wait_printing()
    return self.protocol.command(
        0x50,
        self.password,
        misc.CAST_SIZE['11111'](*misc.int_to_bytes(cash, 5))
    )
income.cmd = 0x50


def outcome(self, cash):
    """
    Выплата.
    """

    self.wait_printing()
    return self.protocol.command(
        0x51,
        self.password,
        misc.CAST_SIZE['11111'](*misc.int_to_bytes(cash, 5))
    )
outcome.cmd = 0x51


def sale(self, item, department_num=0, tax1=0, tax2=0, tax3=0, tax4=0):
    """
    Продажа.
    """

    text, quantity, price = item

    try:
        return self.protocol.command(
            0x80,
            self.password,
            misc.CAST_SIZE['11111'](*misc.int_to_bytes(quantity, 5)),
            misc.CAST_SIZE['11111'](*misc.int_to_bytes(price, 5)),
            misc.CAST_SIZE['1'](department_num),
            misc.CAST_SIZE['1111'](tax1, tax2, tax3, tax4),
            misc.prepare_string(text, self.DEFAULT_MAX_LENGTH)
        )
    except excepts.Error as exc:
        raise excepts.ItemSaleError(exc)
sale.cmd = 0x80


def return_sale(self, item, department_num=0, tax1=0, tax2=0, tax3=0, tax4=0):
    """
    Возврат продажи.
    """

    text, quantity, price = item

    return self.protocol.command(
        0x82,
        self.password,
        misc.CAST_SIZE['11111'](*misc.int_to_bytes(quantity, 5)),
        misc.CAST_SIZE['11111'](*misc.int_to_bytes(price, 5)),
        misc.CAST_SIZE['1'](department_num),
        misc.CAST_SIZE['1111'](tax1, tax2, tax3, tax4),
        misc.prepare_string(text, self.DEFAULT_MAX_LENGTH)
    )
return_sale.cmd = 0x82


def close_check(self,
                cash=0,
                payment_type2=0,
                payment_type3=0,
                payment_type4=0,
                discount_allowance=0,
                tax1=0,
                tax2=0,
                tax3=0,
                tax4=0,
                text=None):
    """
    Закрытие чека.
    """

    try:
        return self.protocol.command(
            0x85,
            self.password,
            misc.CAST_SIZE['11111'](*misc.int_to_bytes(cash, 5)),
            misc.CAST_SIZE['11111'](*misc.int_to_bytes(payment_type2, 5)),
            misc.CAST_SIZE['11111'](*misc.int_to_bytes(payment_type3, 5)),
            misc.CAST_SIZE['11111'](*misc.int_to_bytes(payment_type4, 5)),
            # TODO: проверить скидку/надбавку
            misc.CAST_SIZE['s2'](discount_allowance),
            misc.CAST_SIZE['1111'](tax1, tax2, tax3, tax4),
            misc.prepare_string(text, self.DEFAULT_MAX_LENGTH)
        )
    except excepts.ProtocolError as exc:
        raise excepts.CloseCheckError(exc)
close_check.cmd = 0x85


def discount(self, sum_, tax1=0, tax2=0, tax3=0, tax4=0, text=None):
    """
    Скидка.
    """

    return self.protocol.command(
        0x86,
        self.password,
        misc.CAST_SIZE['11111'](*misc.int_to_bytes(sum_, 5)),
        misc.CAST_SIZE['1111'](tax1, tax2, tax3, tax4),
        misc.prepare_string(text, self.DEFAULT_MAX_LENGTH)
    )
discount.cmd = 0x86


def allowance(self, sum_, tax1=0, tax2=0, tax3=0, tax4=0, text=None):
    """
    Надбавка.
    """

    return self.protocol.command(
        0x87,
        self.password,
        misc.CAST_SIZE['11111'](*misc.int_to_bytes(sum_, 5)),
        misc.CAST_SIZE['1111'](tax1, tax2, tax3, tax4),
        misc.prepare_string(text, self.DEFAULT_MAX_LENGTH)
    )
allowance.cmd = 0x87


def cancel_check(self):
    """
    Аннулирование чека.
    """

    self.wait_printing()
    return self.protocol.command(
        0x88,
        self.password
    )
cancel_check.cmd = 0x88


def repeat(self):
    """
    Повтор документа.
    """

    self.wait_printing()
    return self.protocol.command(
        0x8C,
        self.password
    )
repeat.cmd = 0x8C


def open_check(self, check_type):
    """
    Открыть чек.
    """

    self.wait_printing()
    try:
        return self.protocol.command(
            0x8D,
            self.password,
            misc.CAST_SIZE['1'](check_type)
        )
    except excepts.Error as exc:
        raise excepts.OpenCheckError(exc)
open_check.cmd = 0x8D


def continue_print(self):
    """
    Продолжение печати.
    """

    return self.protocol.command(
        0xB0,
        self.admin_password
    )
continue_print.cmd = 0xB0


def load_graphics(self, line_num, *args):
    """
    Загрузка графики.
    """

    return self.protocol.command(
        0xC0,
        self.password,
        misc.CAST_SIZE['1'](line_num),
        *args
    )
load_graphics.cmd = 0xC0


def print_graphics(self, start_line, end_line):
    """
    Печать графики.
    """

    self.wait_printing()
    return self.protocol.command(
        0xC1,
        self.password,
        misc.CAST_SIZE['1'](start_line),
        misc.CAST_SIZE['1'](end_line)
    )
print_graphics.cmd = 0xC1


def print_barcode(self, num):
    """
    Печать штрих-кода
    """

    self.wait_printing()
    return self.protocol.command(
        0xC2,
        self.password,
        misc.CAST_SIZE['11111'](*misc.int_to_bytes(num, 5))
    )
print_barcode.cmd = 0xC2


def open_shift(self):
    """
    Открыть смену.
    """

    return self.protocol.command(
        0xE0,
        self.password
    )
open_shift.cmd = 0xE0


def model(self):
    """
    Получить тип устройства.
    """

    return self.protocol.command_nopass(
        0xFC
    )
model.cmd = 0xFC


def fs_state(self):
    """
    Запрос статуса ФН.
    """

    return self.protocol.command(
        0xFF01,
        self.admin_password
    )
fs_state.cmd = 0xFF01


def fs_expiration_time(self):
    """
    Запрос срока действия ФН.
    """

    return self.protocol.command(
        0xFF03,
        self.admin_password
    )
fs_expiration_time.cmd = 0xFF03


def fs_cancel_document(self):
    """
    Отменить документ в ФН.
    """

    return self.protocol.command(
        0xFF08,
        self.admin_password
    )
fs_cancel_document.cmd = 0xFF08


def fs_find_document_by_num(self, num):
    """
    Найти фискальный документ по номеру.
    """

    return self.protocol.command(
        0xFF0A,
        self.admin_password,
        misc.CAST_SIZE['4'](num)
    )
fs_find_document_by_num.cmd = 0xFF0A


def fs_open_shift(self):
    """
    Открыть смену в ФН.
    """

    return self.protocol.command(
        0xFF0B,
        self.admin_password
    )
fs_open_shift.cmd = 0xFF0B


def send_tlv_struct(self, tlv_struct):
    """
    Передать произвольную TLV структуру.
    """

    if len(tlv_struct) > misc.TLV_LEN_MAX:
        raise ValueError(
            u'Максимальный размер tlv структуры - {} байт'.format(misc.TLV_LEN_MAX)
        )

    return self.protocol.command(
        0xFF0C,
        self.admin_password,
        tlv_struct
    )
send_tlv_struct.cmd = 0xFF0C


def fs_begin_correction_check(self):
    """
    Начать формирование чека коррекции.
    """

    return self.protocol.command(
        0xFF35,
        self.admin_password
    )
fs_begin_correction_check.cmd = 0xFF35


def fs_correction_check(self, sum_, check_type):
    """
    Сформировать чек коррекции.
    """

    return self.protocol.command(
        0xFF36,
        self.admin_password,
        misc.CAST_SIZE['11111'](*misc.int_to_bytes(sum_, 5)),
        misc.CAST_SIZE['1'](check_type)
    )
fs_correction_check.cmd = 0xFF36


def fs_calculation_state_report(self):
    """
    Сформировать отчёт о состоянии расчётов.
    """

    return self.protocol.command(
        0xFF38,
        self.admin_password
    )
fs_calculation_state_report.cmd = 0xFF38


def fs_info_exchange(self):
    """
    Получить статус информационного обмена.
    """

    return self.protocol.command(
        0xFF39,
        self.admin_password
    )
fs_info_exchange.cmd = 0xFF39


def fs_unconfirmed_document_count(self):
    """
    Запрос количества ФД на которые нет квитанции.
    """

    return self.protocol.command(
        0xFF3F,
        self.admin_password
    )
fs_unconfirmed_document_count.cmd = 0xFF3F


def fs_shift_params(self):
    """
    Запрос параметров текущей смены.
    """

    return self.protocol.command(
        0xFF40,
        self.admin_password
    )
fs_shift_params.cmd = 0xFF40


def fs_begin_open_shift(self):
    """
    Начать открытие смены.
    """

    return self.protocol.command(
        0xFF41,
        self.admin_password
    )
fs_begin_open_shift.cmd = 0xFF41


def fs_begin_close_shift(self):
    """
    Начать закрытие смены.
    """

    return self.protocol.command(
        0xFF42,
        self.admin_password
    )
fs_begin_close_shift.cmd = 0xFF42


def fs_close_shift(self):
    """
    Закрыть смену в ФН.
    """

    return self.protocol.command(
        0xFF43,
        self.admin_password
    )
fs_close_shift.cmd = 0xFF43


def wait_printing(self):
    """
    Метод ожидания окончания печати документа.
    """

    while True:
        time.sleep(self.WAIT_TIME)

        state_ = self.state()
        mode = state_[u'Режим ФР']
        submode = state_[u'Подрежим ФР']

        if mode.num == 12:
            continue

        if submode.state == 0:
            return
        if submode.state == 3:
            self.continue_print()
print_string.depends = (wait_printing, )
test_start.depends = (wait_printing, )
cut.depends = (wait_printing, )
feed.depends = (wait_printing, )
test_stop.depends = (wait_printing, )
x_report.depends = (wait_printing, )
z_report.depends = (wait_printing, )
income.depends = (wait_printing, )
outcome.depends = (wait_printing, )
cancel_check.depends = (wait_printing, )
repeat.depends = (wait_printing, )
open_check.depends = (wait_printing, )
print_graphics.depends = (wait_printing, )
print_barcode.depends = (wait_printing, )


module_ = sys.modules[__name__]
FUNCTIONS = {
    function_.cmd if hasattr(function_, 'cmd') else name: function_
    for name, function_ in inspect.getmembers(module_, inspect.isfunction)
    if function_.__module__ == module_.__name__
}


class SupportedCommands(type):
    def __new__(mcs, classname, supers, attributedict):
        command_nums = attributedict.get('SUPPORTED_COMMANDS', ())

        def relative_gen(cmd):
            if hasattr(cmd, 'depends'):
                for dpc in cmd.depends:
                    relative_gen(dpc)
                    if dpc not in attributedict.values():
                        yield dpc

            if hasattr(cmd, 'required'):
                if all(rqc in attributedict.values() for rqc in cmd.required):
                    yield cmd

            if hasattr(cmd, 'related'):
                for rlc in cmd.related:
                    relative_gen(rlc)
                    if rlc not in attributedict.values():
                        yield rlc

        for cn in command_nums:
            command = FUNCTIONS[cn]
            attributedict[command.__name__] = command

            for c in relative_gen(command):
                attributedict[c.__name__] = c

        return super(SupportedCommands, mcs).__new__(mcs, classname, supers, attributedict)
